from __future__ import annotations

from typing import Iterable, Iterator, Sequence

from numba import jit
from numpy import log

from pyfracgen.common import CanvasBounded, Result
from pyfracgen.types import Bound, Lattice, UpdateFunc
from pyfracgen.updaters.funcs import power


@jit  # type: ignore[misc]
def _julia_paint(
    c: float,
    xvals: Sequence[float],
    yvals: Sequence[float],
    lattice: Lattice,
    update_func: UpdateFunc,
    maxiter: int,
    horizon: float,
    log_smooth: bool,
) -> None:

    logh = log(log(horizon)) / log(2)
    for iy, yval in enumerate(yvals):
        for ix, xval in enumerate(xvals):
            z = xval + 1j * yval
            for it in range(maxiter):
                az = abs(z)
                if az > horizon:
                    if log_smooth:
                        lattice[iy, ix] = it - log(log(az)) / log(2) + logh
                    else:
                        lattice[iy, ix] = it
                    break
                z = update_func(z, c)


class Julia(CanvasBounded):
    def paint(
        self,
        c: complex,
        update_func: UpdateFunc,
        maxiter: int,
        horizon: float,
        log_smooth: bool,
    ) -> None:

        _julia_paint(
            c,
            self.xvals,
            self.yvals,
            self.lattice,
            update_func,
            maxiter,
            horizon,
            log_smooth,
        )


def julia(
    cvals: Iterable[complex],
    xbound: Bound,
    ybound: Bound,
    update_func: UpdateFunc = power,
    width: int = 5,
    height: int = 4,
    dpi: int = 300,
    maxiter: int = 100,
    horizon: float = 2.0**40,
    log_smooth: bool = True,
) -> Iterator[Result]:

    for c in cvals:
        canvas = Julia(width, height, dpi, xbound, ybound)
        canvas.paint(c, update_func, maxiter, horizon, log_smooth)
        yield canvas.result



from .buddhabrot import buddhabrot
from .common import Canvas, Result
from .images import images
from .julia import julia
from .lyapunov import lyapunov
from .mandelbrot import mandelbrot
from .randomwalk import construct_moves, randomwalk
from .updaters import funcs

__all__ = [
    "buddhabrot",
    "images",
    "julia",
    "julia_series",
    "lyapunov",
    "mandelbrot",
    "randomwalk",
    "construct_moves",
    "Result",
    "Canvas",
    "funcs",
]
__version__ = "0.0.11"



from __future__ import annotations

from typing import Callable, NewType, Tuple

from nptyping import Complex128, Float64, Int64, NDArray, Shape

Width = NewType("Width", int)
Height = NewType("Height", int)
Depth = NewType("Depth", int)

Lattice = NDArray[Shape["Height, Width"], Float64]
Lattice3D = NDArray[Shape["Height, Width, Depth"], Float64]
Moves3D = NDArray[Shape["*, 3"], Int64]
Boxes = NDArray[Shape["*, 2"], Float64]
ComplexSequence = NDArray[Shape["*"], Complex128]

Bound = Tuple[float, float]
UpdateFunc = Callable[[complex, complex], complex]



from __future__ import annotations

from typing import Sequence

import numpy as np
from numba import jit

from pyfracgen.common import CanvasBounded, Result
from pyfracgen.types import Bound, Lattice


@jit  # type: ignore[misc]
def _lyapunov_paint(
    string: str,
    xvals: Sequence[float],
    yvals: Sequence[float],
    lattice: Lattice,
    ninit: int,
    niter: int,
) -> None:

    instructions = [0 if s == "A" else 1 for s in string]
    length = len(instructions)
    for iy, yval in enumerate(yvals):
        for ix, xval in enumerate(xvals):
            coord = [xval, yval]
            count = 0
            x = 0.5
            for _ in range(ninit):
                rn = coord[instructions[count % length]]
                x = (rn * x) * (1 - x)
                count += 1
            lamd = 0.0
            for _ in range(niter):
                rn = coord[instructions[count % length]]
                x = (rn * x) * (1 - x)
                lamd += np.log(np.abs(rn - 2 * rn * x))
                count += 1
            lattice[iy, ix] += lamd / niter


class Lyapunov(CanvasBounded):
    def paint(
        self,
        string: str,
        ninit: int,
        niter: int,
    ) -> None:

        _lyapunov_paint(string, self.xvals, self.yvals, self.lattice, ninit, niter)


def lyapunov(
    string: str,
    xbound: Bound,
    ybound: Bound,
    ninit: int = 200,
    niter: int = 800,
    width: int = 5,
    height: int = 4,
    dpi: int = 300,
) -> Result:

    canvas = Lyapunov(width, height, dpi, xbound, ybound)
    canvas.paint(string, ninit, niter)
    return canvas.result



from __future__ import annotations

from typing import Sequence

from numba import jit
from numpy import log

from pyfracgen.common import CanvasBounded, Result
from pyfracgen.types import Bound, Lattice, UpdateFunc
from pyfracgen.updaters.funcs import power


@jit  # type: ignore[misc]
def _mandelbrot_paint(
    xvals: Sequence[float],
    yvals: Sequence[float],
    lattice: Lattice,
    update_func: UpdateFunc,
    maxiter: int,
    horizon: float,
    log_smooth: bool,
) -> None:

    logh = log(log(horizon)) / log(2)
    for iy, yval in enumerate(yvals):
        for ix, xval in enumerate(xvals):
            c = xval + 1j * yval
            z = c
            for it in range(maxiter):
                az = abs(z)
                if az > horizon:
                    if log_smooth:
                        lattice[iy, ix] = it - log(log(az)) / log(2) + logh
                    else:
                        lattice[iy, ix] = it
                    break
                z = update_func(z, c)


class Mandelbrot(CanvasBounded):
    def paint(
        self, update_func: UpdateFunc, maxiter: int, horizon: float, log_smooth: bool
    ) -> None:

        _mandelbrot_paint(
            self.xvals,
            self.yvals,
            self.lattice,
            update_func,
            maxiter,
            horizon,
            log_smooth,
        )


def mandelbrot(
    xbound: Bound,
    ybound: Bound,
    update_func: UpdateFunc = power,
    width: int = 5,
    height: int = 4,
    dpi: int = 300,
    maxiter: int = 1000,
    horizon: float = 2.0**40,
    log_smooth: bool = True,
) -> Result:

    canvas = Mandelbrot(width, height, dpi, xbound, ybound)
    canvas.paint(update_func, maxiter, horizon, log_smooth)
    return canvas.result
